import numpy as np
import matplotlib.pyplot as plt

Notes = '''
RK for systems (of m equations):
    Note: This function was heavily influenced by the style of Dr. Andasari (BU)
    website: http://people.bu.edu/andasari/courses/numericalpython/python.html

    Function Name: RK4_sys

    Inputs: func: The defining functions written in the following style
                func(t, y):
                    dy = [0, 0, ..., 0]
                    dy[0] = u1' = f1(t, y[j]); j in 1,2,..., m
                    dy[1] = u2' = f2(t, y[j]); j in 1,2,..., m
                    .
                    .
                    .
                    dy[m] = um' = fm(t, y[j]); j in 1,2,..., m
                    return dy
            yinit: Array of initial conditions [u1(a), u2(a), ..., um(a)]

            x_range: The array [a, b]

            h: stepsize

    Outputs:
            [tsol, ysol]:
                ysol: An array of arrays of solutions [u1, u2, ..., um]
                tsol: Time axis
'''


def RK(f, yinit, a, b, h, order = 4, plot = 0):
    if(order == 3):
        [xsol, solutions] = RK3(f, yinit, a, b, h,  plot = plot)
        return [xsol, solutions]
    if(order == 4):
        [xsol, solutions] = RK4(f, yinit, a, b, h,  plot = plot)
        return [xsol, solutions]
    if(order == 6):
        [xsol, solutions] = RK6(f, yinit, a, b, h,  plot = plot)
        return [xsol, solutions]
    else:
        print("Orders supported: 3, 4 and 6")

    return 0


#---------------------------RK 3rd order method--------------------------------
def RK3(f, yinit, a, b, h,  plot = 0):

    m = len(yinit)
    n = int((b - a)/h)

    x = a
    y = yinit

    # Containers for solutions
    xsol = np.empty(0)
    xsol = np.append(xsol, x)

    ysol = np.empty(0)
    ysol = np.append(ysol, y)

    for i in range(n):
        k1 = f(x, y)
        yp2 = y + k1*(h/2)
        k2 = f(x+h/2, yp2)
        yp3 = y - k1*h + 2*k2*h
        k3 = f(x+h, yp3)
        for j in range(m):
            y[j] = y[j] + (h/6)*(k1[j] + 4*k2[j] + k3[j])
        x = x + h
        xsol = np.append(xsol, x)
        for r in range(len(y)):
            ysol = np.append(ysol, y[r])

    solutions = []
    for q in range(m):
        temp = ysol[q::m]
        solutions.append(temp)
        temp = []
    if(plot == 1):
        for i in range(len(solutions)):
            plt.plot(xsol, solutions[i], label = "u_%s"%(i+1))
        plt.legend()
        plt.show()
    return [xsol, solutions]


#--------------------- RK 4th order method ---------------------------

def RK4(f, yinit, a, b, h,  plot = 0):

    m = len(yinit)
    n = int((b - a)/h)

    x = a
    y = yinit

    # Containers for solutions
    xsol = np.empty(0)
    xsol = np.append(xsol, x)

    ysol = np.empty(0)
    ysol = np.append(ysol, y)

    for i in range(n):
        k1 = f(x, y)
        yp2 = y + k1*(h/2)
        k2 = f(x+h/2, yp2)
        yp3 = y + k2*(h/2)
        k3 = f(x+h/2, yp3)
        yp4 = y + k3*h
        k4 = f(x+h, yp4)
        for j in range(m):
            y[j] = y[j] + (h/6)*(k1[j] + 2*k2[j] + 2*k3[j] + k4[j])
        x = x + h
        xsol = np.append(xsol, x)
        for r in range(len(y)):
            ysol = np.append(ysol, y[r])

    solutions = []
    for q in range(m):
        temp = ysol[q::m]
        solutions.append(temp)
        temp = []
    if(plot == 1):
        for i in range(len(solutions)):
            plt.plot(xsol, solutions[i], label = "u_%s"%(i+1))
        plt.legend()
        plt.show()
    return [xsol, solutions]

#------------------------ RK6th order -------------------------------

def RK6(f, yinit, a, b, h,  plot = 0):

    m = len(yinit)
    n = int((b - a)/h)

    x = a
    y = yinit

    # Containers for solutions
    xsol = np.empty(0)
    xsol = np.append(xsol, x)

    ysol = np.empty(0)
    ysol = np.append(ysol, y)

    for i in range(n):
        k1 = h*f(x, y)
        yp2 = y + k1
        k2 = h*f(x+h, yp2)
        yp3 = y + (1/8)*(3*k1 + k2)
        k3 = h*f(x+h/2, yp3)
        yp4 = y + (1/27)*(8*k1 + 2*k2 + 8*k3)
        k4 = h*f(x+ (2*h/3), yp4)
        yp5 = y + (1/392)*(3*(3*np.sqrt(21) -7)*k1 - 8*(7-np.sqrt(21))*k2 + 48*(7-np.sqrt(21))*k3 - 3*(21 - np.sqrt(21))*k4 )
        k5 = h*f(x + (h/14)*(7-np.sqrt(21)), yp5)
        yp6 = y + (1/1960)*(-5*(231 + 51*np.sqrt(21))*k1 - 40*(7 + np.sqrt(21))*k2 - 320*np.sqrt(21)*k3 + 3*(21 + 121*np.sqrt(21))*k4 + 392*(6 + np.sqrt(21))*k5)
        k6 = h*f(x + (h/14)*(7+np.sqrt(21)), yp6)
        yp7 = y + (1/180)*(15*(22 + 7*np.sqrt(21))*k1 + 120*k2 + 40*(7*np.sqrt(21) - 5)*k3 - 63*(3*np.sqrt(21) -2)*k4 - 14*(49 + 9*np.sqrt(21))*k5 + 70*(7-np.sqrt(21))*k6)
        k7 = h*f(x+h, yp7)

        for j in range(m):
            y[j] = y[j] + (1/180)*(9*k1[j] + 64*k2[j] + 49*k5[j] + 49*k6[j] + 9*k7[j])
        x = x + h
        xsol = np.append(xsol, x)
        for r in range(len(y)):
            ysol = np.append(ysol, y[r])

    solutions = []
    for q in range(m):
        temp = ysol[q::m]
        solutions.append(temp)
        temp = []
    if(plot == 1):
        for i in range(len(solutions)):
            plt.plot(xsol, solutions[i], label = "u_%s"%(i+1))
        plt.legend()
        plt.show()
    return [xsol, solutions]



import numpy as np


def gaussian_elim(n, A, b, ver = 0):
    if (ver == 0):
        sol = gebs(n, A, b)
    if(ver == 1):
        sol = gebs_2(n, A, b)
    if(ver == 2):
        sol = gebs_3(n, A, b)

    return sol

#Gaussian Elimination and Backward Substitution
def gebs(n, A, b):
    Ab = np.hstack([A, b.reshape(-1, 1)])
    x = np.zeros(n)
    #step 1
    for i in range(n):
        #step 2
        p = 99999
        for q in range(i, n):
            if(Ab[q][i] !=0):
                p = q
                break
        if(p == 99999):
            print("No unique solution exists!")
            return 0
        #step 3
        if(p != i):
            print("swapping rows %s and %d!" % (p, i))
            Ab[[p, i]] = Ab[[i, p]]
        #step 4
        for j in range(i+1, n):
            #step 5
            m_ji = Ab[j][i]/Ab[i][i]
            #step 6
            Ab[j] = Ab[j] - m_ji*Ab[i]
    if(Ab[n-1][n-1] == 0):
        print("No unique solution exists!")
        return 0
    x[n-1] = Ab[n-1][n]/Ab[n-1][n-1]
    for r in range(n-2, -1, -1):
        temp_sum = 0
        for j in range(r+1, n):
            temp_sum += Ab[r][j]*x[j]
        x[r] = (Ab[r][n] - temp_sum)/Ab[r][r]

    return x

#-------------------------------------------------------------------------------

#Algorithm 6.2
#Gaussian Elimination and Backward Substitution with partial pivoting
def gebs_2(n, A, b):
    Ab = np.hstack([A, b.reshape(-1, 1)])
    x = np.zeros(n)
    nrow = np.zeros(n)
    #step 1
    for i in range(n):
        nrow[i] = i
    #step 2
    for i in range(n-1):
        #step 3
        p = 99999
        max = -1
        for q in range(i,n):
            if(np.abs(Ab[int(nrow[q])][i]) > max):
                max = np.abs(Ab[int(nrow[q])][i])
        for d in range(i,n):
            if(np.abs(Ab[int(nrow[d])][i]) == max):
                p = d
                break
        if(Ab[int(nrow[p])][i] == 0):
            print("no unique solution exists")
            return 0
        if(nrow[i] != nrow[p]):
            print("exchanging rows %u and %a"%(int(nrow[i] +1), int(nrow[p] +1)))
            ncopy = nrow[i]
            nrow[i] = nrow[p]
            nrow[p] = ncopy
        for j in range(i+1, n):
            m_ji = Ab[int(nrow[j])][i]/Ab[int(nrow[i])][i]
            Ab[int(nrow[j])] = Ab[int(nrow[j])] - m_ji*Ab[int(nrow[i])]
        if(Ab[int(nrow[n-1])][n-1] == 0):
            print("no unique solution exists")
            return 0

    x[n-1] = Ab[int(nrow[n-1])][n]/Ab[int(nrow[n-1])][n-1]
    for r in range(n-2, -1, -1):
        temp_sum = 0
        for j in range(r+1, n):
            temp_sum += Ab[int(nrow[r])][j]*x[j]
        x[r] = (Ab[int(nrow[r])][n] - temp_sum)/Ab[int(nrow[r])][r]
    return x

#-------------------------------------------------------------------------------

#Algotithm 6.3
#Gaussian Elimination and Backward Substitution with scaled pivoting
def gebs_3(n, A, b):
    Ab = np.hstack([A, b.reshape(-1, 1)])
    s = np.zeros(n)
    x = np.zeros(n)
    nrow = np.zeros(n)
    #step 1
    max = 0
    for i in range(n):
        for j in range(n):
            if(np.abs(Ab[i][j]) > max):
                max = np.abs(Ab[i][j])
        s[i] = max
        if(s[i] == 0):
            print("No solution exists!")
            return 0
        else:
            nrow[i] = i
    #step 2
    for i in range(n-1):
        #step 3
        p = 99999
        max = -1
        for q in range(i,n):
            if(np.abs(Ab[int(nrow[q])][i])/s[int(nrow[q])] > max):
                max = np.abs(Ab[int(nrow[q])][i])/s[int(nrow[q])]
        for d in range(i,n):
            if(np.abs(Ab[int(nrow[d])][i])/s[int(nrow[d])] == max):
                p = d
                break
        if(Ab[int(nrow[p])][i] == 0):
            print("no unique solution exists")
            return 0
        if(nrow[i] != nrow[p]):
            print("exchanging rows %u and %a"%(int(nrow[i] +1), int(nrow[p] +1)))
            ncopy = nrow[i]
            nrow[i] = nrow[p]
            nrow[p] = ncopy
        for j in range(i+1, n):
            m_ji = Ab[int(nrow[j])][i]/Ab[int(nrow[i])][i]
            Ab[int(nrow[j])] = Ab[int(nrow[j])] - m_ji*Ab[int(nrow[i])]
        if(Ab[int(nrow[n-1])][n-1] == 0):
            print("no unique solution exists")
            return 0

    x[n-1] = Ab[int(nrow[n-1])][n]/Ab[int(nrow[n-1])][n-1]
    for r in range(n-2, -1, -1):
        temp_sum = 0
        for j in range(r+1, n):
            temp_sum += Ab[int(nrow[r])][j]*x[j]
        x[r] = (Ab[int(nrow[r])][n] - temp_sum)/Ab[int(nrow[r])][r]
    return x



#__init__.py
import numpy as np

from .Adam_Bashforth import *
from .Euler import *
from .Predictor_Corrector import *
from .RK import *
from .RKF import *
from .Taylor import *
from .Modified_Euler import *
from .help import *



from .post_processing import *
from .post_processing import help as help_general
from .matrix_methods.help import dict2
from .matrix_methods.help import help as help_matrix
from .ODE_solvers.help import dict1
from .ODE_solvers.help import help as help_ode

def help(kw = ""):
    help_general(kw)
    help_matrix(kw)
    help_ode(kw)
    return 0



import numpy as np
import matplotlib.pyplot as plt

Notes = '''
Predictor Corrector scheme (Predictor: Adam Bashforth; Corrector: Adam-Moulton)
    Function Name: predict_correct

    Inputs: f: defining function ( from y' = f(t,y) )
            t: Initial time array [a]
            y: Initial solution array [alpha]
            pred: # steps for predictor (the 'm' in Adam Bashforth m-step)
            corr: # steps for corrector (the 'm' in Adam Moulton m-step)
            alpha: Initial Condition
            a: Initial time
            b: Final time
            h: Step size

    Outputs: y: Solution array


Adams variable stepsize predictor corrector

    Note: This method uses 4 step Adam bashforth and 3 step Adam Moulton
    as Predictor and Corrector Respectively.

    Function Name: Adam_PC_variable_step

    Inputs: f: defining function ( from y' = f(t,y) )
            t: Initial time array [a]
            y: Initial solution array [alpha]
            a: Initial time
            b: Final time
            tol: Tolerance value
            hmax: maximum stepsize
            hmin: minimum stepsize

    Outputs: Index: Array of indices (as time array is not of uniform step size)
             t: Time array
             y: Solution array
             h: Step size array

'''


#--------------------- RK 4th order method ---------------------------

def RK4_method(f, a, b, t, y, h, N=0):
    if(N == 0):
        N = int((b-a)/h)

    for i in range(N):
        k1 = h*f(t[-1], y[-1])
        yp2 = y[-1] + k1/2
        k2 = h*f(t[-1]+h/2, yp2)
        yp3 = y[-1] + k2/2
        k3 = h*f(t[-1]+h/2, yp3)
        yp4 = y[-1] + k3
        k4 = h*f(t[-1]+h, yp4)
        val = (1/6)*(k1 + 2*k2 + 2*k3 + k4)
        y.append(y[-1] + val)
        t.append(t[-1] + h)
    return t, y


#----------------- Adam Bashforth explicit methods --------------------

#Adam Bashforth handler

def adam_bashforth(f, a, b, alpha, h, type, N = 0):
    t = [a]
    y = [alpha]
    t, y = RK4_method(f, a, b, t, y, h, type - 1)
    if(N == 0):
        N = int(((b-a)/h)- (type-1))

    if(type == 2):
        y = abf2(f, y, t, h, N)
    if(type == 3):
        y = abf3(f, y, t, h, N)
    if(type == 4):
        y = abf4(f, y, t, h, N)
    if(type == 5):
        y = abf5(f, y, t, h, N)

    return y

#2 step Adam Bashforth
def abf2(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/2)*(3*f(t[-1], y[-1]) - f(t[-2], y[-2])))
        t.append(t[-1] + h)
    return y

#3 step Adam Bashforth
def abf3(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/12)*(23*f(t[-1], y[-1]) - 16*f(t[-2], y[-2]) \
                                + 5*f(t[-3], y[-3])))
        t.append(t[-1] + h)
    return y

#4 step Adam Bashforth
def abf4(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/24)*(55*f(t[-1], y[-1]) - 59*f(t[-2], y[-2]) \
                                    + 37*f(t[-3], y[-3]) - 9*f(t[-4], y[-4])))
        t.append(t[-1] + h)
    return y

#5 step Adam bashforth
def abf5(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/720)*(1901*f(t[-1], y[-1]) - 2774*f(t[-2], y[-2]) \
                                + 2616*f(t[-3], y[-3]) - 1274*f(t[-4], y[-4]) \
                                +251*f(t[-5], y[-5])))
        t.append(t[-1] + h)
    return y


#----------------------------------------------------------------------


#------ Adam Moulton Implicit method (for predictor corrector)----------

#Adam Moulton method handler
def adam_moulton(f, t, y, h, type, wp):

    if(type == 2):
        t, y = abm2(f, y, t, h, 1, wp)
    if(type == 3):
        t, y = abm3(f, y, t, h, 1, wp)
    if(type == 4):
        t, y = abm4(f, y, t, h, 1, wp)

    return t, y

#2 step Adam Moulton
def abm2(f, y, t, h, N, wp):
    for i in range(N):
        y.append(y[-1] + (h/12)*(5*f(t[-1] + h, wp) + 8*f(t[-1], y[-1]) \
                            - f(t[-2], y[-2])))
        t.append(t[-1] + h)
    return t, y

#3 step Adam Moulton
def abm3(f, y, t, h, N, wp):
    for i in range(N):
        y.append(y[-1] + (h/24)*(9*f(t[-1]+h, wp) + 19*f(t[-1], y[-1]) \
                                - 5*f(t[-2], y[-2]) + f(t[-3], y[-3])))
        t.append(t[-1] + h)
    return t, y

#4 step Adam Moulton
def abm4(f, y, t, h, N, wp):
    for i in range(N):
        y.append(y[-1] + (h/720)*(251*f(t[-1]+h, wp) + 646*f(t[-1], y[-1]) \
                                - 264*f(t[-2], y[-2]) + 106*f(t[-3], y[-3]) \
                                -19*f(t[-4], y[-4])))
        t.append(t[-1] + h)
    return t, y


#--------------- Predictor Corrector -------------------------

def predict_correct(f, alpha, a, b, pred, corr,  var_step = 0, plot = 0):
    t = [a]
    y = [alpha]
    if(var_step == 1):
        print("Using pred = 4, corr = 3")
        hmax = float(input("Enter maximum threshold for step-size: "))
        hmin = float(input("Enter minimum threshold for step-size: "))
        tol = float(input("Enter tolerance value: "))
        index, t, y, step = Adam_PC_variable_step(f, t, y, a, b, hmax, hmin, tol)
        if(plot == 1):
            plt.plot(t, y, label = "Variable h PC")
            plt.legend()
            plt.show()
        return [t, y, step]

    h = float(input("Enter the step-size: "))
    N = int(((b-a)/h)- (pred-1))
    t, y = RK4_method(f, a, b, t, y, h, pred - 1)
    for i in range(N):
        #the following predictor is very inefficient.
        #But it works for smaller problems.
        #It can, and will be optimized in further updates

        wp_arr = adam_bashforth(f, a, b, alpha, h, pred, i+1)
        wp = wp_arr[-1]
        t, y = adam_moulton(f, t, y, h, corr, wp)
    if(plot == 1):
        plt.plot(t, y, label = "Predictor Corrector Solution")
        plt.legend()
        plt.show()

    return [t, y]

#-------------------------------------------------------------

#------------------------------------------------------------
#Adam's Predictor Corrector method with variable step size

def Adam_PC_variable_step(f, t, y, a, b, tol, hmax, hmin):
    def RK4(h, v0, x0):
        x = [x0, 0, 0, 0]
        v = [v0, 0, 0, 0]
        for j in range(1,4):
            k1 = h*f(x[j-1], v[j-1])
            k2 = h*f(x[j-1] + h/2, v[j-1] + k1/2)
            k3 = h*f(x[j-1] + h/2, v[j-1] + k2/2)
            k4 = h*f(x[j-1] + h, v[j-1] + k3)
            v[j] = v[j-1] + (1/6)*(k1 + 2*k2 + 2*k3 + k4)
            x[j] = x[j-1] + h

        return x[1:], v[1:]

    index = [0]
    step = [hmax]
    t0 = a
    w0 = y[0]
    h = hmax
    flag = 1
    last = 0

    tpr, ypr = RK4(h, w0, t0)
    for i in range(len(tpr)):
        t.append(tpr[i])
        y.append(ypr[i])


    nflag = 1
    i = 4
    time = t[i-1] + h

    while(flag == 1):
        wp = y[i-1] + (h/24)*(55*f(t[i-1], y[i-1]) - 59*f(t[i-2], y[i-2]) \
                                    + 37*f(t[i-3], y[i-3]) - 9*f(t[i-4], y[i-4]))
        wc = y[i-1] + (h/24)*(9*f(time, wp) + 19*f(t[i-1], y[i-1]) \
                                - 5*f(t[i-2], y[i-2]) + f(t[i-3], y[i-3]))
        sigma = (19/(270*h))*np.abs(wc - wp)

        if(sigma < tol):

            y.append(wc)
            t.append(time)

            if(nflag == 1):
                for j in range(i-3, i+1):
                    step.append(h)
                    index.append(j)
            else:
                step.append(h)
                index.append(i)

            if(last == 1):
                flag = 0
            else:
                i = i + 1
                nflag = 0

                if(sigma < 0.1*tol or t[i-1] + h > b):

                    q = (tol/(2*sigma))**(1/4)

                    if(q > 4):
                        h = 4*h
                    else:
                        h = q*h

                    if(h > hmax):
                        h = hmax

                    if(t[i-1] + 4*h > b):
                        h = (b - t[i-1])/4
                        last = 1

                    temp1 = []
                    temp2 = []
                    temp1, temp2 = RK4(h, y[i-1], t[i-1])
                    for v in range(len(temp1)):
                        t.append(temp1[v])
                        y.append(temp2[v])
                    nflag = 1
                    i = i+3

        else:

            q = (tol/(2*sigma))**(1/4)


            if(q < 0.1):
                h = 0.1*h
            else:
                h = q*h
            if(h < hmin):
                flag = 0
                print("min h exceeded")
            else:
                if(nflag == 1):
                    i = i-3
                y = y[:i]
                t = t[:i]
                temp3 = []
                temp4 = []
                temp3, temp4 = RK4(h, y[i-1], t[i-1])

                for v in range(len(temp3)):
                    t.append(temp3[v])
                    y.append(temp4[v])
                i = i+3
                nflag = 1
        time = t[i-1] + h
    return index, t, y, step



import numpy as np

def Matmul(A, B):
    if(len(A[0]) != len(B)):
        print("Matrices not compatible")
    res = np.zeros((len(A), len(B[0])))
    # iterate through rows of X
    for i in range(len(A)):
       # iterate through columns of Y
       for j in range(len(B[0])):
           # iterate through rows of Y
           for k in range(len(B)):
               res[i][j] += A[i][k] * B[k][j]
    return res


def determinant(A):
    n = len(A)
    AM = np.copy(A)
    def Upper_T(n, AM):
        for diag in range(n):
            for i in range(diag +1, n):
                if(AM[diag][diag] == 0):
                    AM[diag][diag] = 1*10**(-16)
                temp = AM[i][diag]/AM[diag][diag]
                for j in range(n):
                    AM[i][j] = AM[i][j] - temp*AM[diag][j]
        return AM
    res = Upper_T(n, AM)
    
    prod = 1.0
    for i in range(n):
        prod *= AM[i][i]

    return prod



import numpy as np
import matplotlib.pyplot as plt

'''
Adam Bashforth m step method:
    Function Name: adam_bashforth

    Inputs: f: defining function ( from y' = f(t,y) )
            a: starting Time
            b: ending Time
            alpha: Initial Condition
            h: stepsize
            type: # of steps (the 'm' in m-step scheme)
                                m = 2,3,4 or 5
            N = (optional) to override the number of iterations
                default: N = 0 for normal Adam Bashforth explicit
                    reset N to any finite number i to get i iterations.
                        (i >= type-1, for obvious reasons)
            plot: Binary input
                Default plot = 0: No plotting
                        plot = 1: plots the solution

    Outputs: [t, y]: where
                t: time axis
                y: solution array
'''

#--------------------- RK 4th order method ---------------------------

def RK4_method(f, a, b, t, y, h, N=0):
    if(N == 0):
        N = int((b-a)/h)

    for i in range(N):
        k1 = h*f(t[-1], y[-1])
        yp2 = y[-1] + k1/2
        k2 = h*f(t[-1]+h/2, yp2)
        yp3 = y[-1] + k2/2
        k3 = h*f(t[-1]+h/2, yp3)
        yp4 = y[-1] + k3
        k4 = h*f(t[-1]+h, yp4)
        val = (1/6)*(k1 + 2*k2 + 2*k3 + k4)
        y.append(y[-1] + val)
        t.append(t[-1] + h)
    return t, y


#----------------- Adam Bashforth explicit methods --------------------

#Adam Bashforth handler
def adam_bashforth(f, a, b, alpha, h, type, N = 0, plot = 0):
    t = [a]
    y = [alpha]
    t, y = RK4_method(f, a, b, t, y, h, type - 1)
    if(N == 0):
        N = int(((b-a)/h) - (type-1))

    if(type == 2):
        t, y = abf2(f, y, t, h, N)
        if(plot == 1):
            plt.plot(t, y, label = "AB %s step solution"%type)
            plt.legend()
            plt.show()

    if(type == 3):
        t, y = abf3(f, y, t, h, N)
        if(plot == 1):
            plt.plot(t, y, label = "AB %s step solution"%type)
            plt.legend()
            plt.show()

    if(type == 4):
        t, y = abf4(f, y, t, h, N)
        if(plot == 1):
            plt.plot(t, y, label = "AB %s step solution"%type)
            plt.legend()
            plt.show()

    if(type == 5):
        t, y = abf5(f, y, t, h, N)
        if(plot == 1):
            plt.plot(t, y, label = "AB %s step solution"%type)
            plt.legend()
            plt.show()

    return [t, y]

#2 step Adam Bashforth
def abf2(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/2)*(3*f(t[-1], y[-1]) - f(t[-2], y[-2])))
        t.append(t[-1] + h)
    return t, y

#3 step Adam Bashforth
def abf3(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/12)*(23*f(t[-1], y[-1]) - 16*f(t[-2], y[-2]) \
                                + 5*f(t[-3], y[-3])))
        t.append(t[-1] + h)
    return t, y

#4 step Adam Bashforth
def abf4(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/24)*(55*f(t[-1], y[-1]) - 59*f(t[-2], y[-2]) \
                                    + 37*f(t[-3], y[-3]) - 9*f(t[-4], y[-4])))
        t.append(t[-1] + h)
    return t, y

#5 step Adam bashforth
def abf5(f, y, t, h, N):
    for i in range(N):
        y.append(y[-1] + (h/720)*(1901*f(t[-1], y[-1]) - 2774*f(t[-2], y[-2]) \
                                + 2616*f(t[-3], y[-3]) - 1274*f(t[-4], y[-4]) \
                                +251*f(t[-5], y[-5])))
        t.append(t[-1] + h)
    return t, y



import numpy as np
import matplotlib.pyplot as plt

Notes_euler = '''
Euler's method:
    Function Name: Euler

    Inputs: f: defining function ( from y' = f(t,y) )
            a: starting Time
            b: ending Time
            y0: Initial Condition
            h: stepsize
            plot: Binary input
                Default plot = 0: No plotting
                        plot = 1: plots the solution

    Outputs: [t, y] where
        y: solution array
        t: time axis
'''


def Euler(f, a, b, y0, h, plot = 0):
    t = [a]
    y = [y0]
    N = int((b-a)/h)
    for i in range(N):
        y.append(y[-1] + h*f(t[-1], y[-1])) #euler update
        t.append(t[-1] + h) #time update

    if(plot == 1):
        plt.plot(t, y, label = "solution")
        plt.legend()
        plt.show()

    return [t, y]



import numpy as np
import matplotlib.pyplot as plt

Notes = '''
nth order Taylor method (n = 2, 3, 4, 5):
    Function Name: Taylor

    Inputs: y0: Initial condition
            h = step size
            a: starting Time
            b: ending Time
            plot: Binary input
                Default plot = 0: No plotting
                        plot = 1: plots the solution
            F = f (the defining function y' = f(t, y))
            Fpr = derivative of f
            F2pr = second derivative of f
            F3pr = third derivative of f
            F4pr = fourth derivative of f

            Specify only to requirement.
                -------Example 1:

                Suppose we want to use 2nd order Taylor. Then define
                functions f and f' then call function as follows

                def Taylor(y0, h, a, b, plot = 0, F=f, Fpr=f'):

                -------Example 2:

                Suppose we want to use 3rd order Taylor. Then define
                functions f, f' and f'' then call function as follows

                def Taylor(y0, h, a, b, plot = 0, F=f, Fpr=f', F2pr = f''):


    Output:  [t, y] where
                y: solution array
                t: time array

'''
def temp_func(t, y):
    return 0

#Computing T^(n)
def T_n(t, y, h, F, Fpr, F2pr, F3pr, F4pr):
    if(F==0):
        F = temp_func
    if(Fpr==0):
        Fpr = temp_func
    if(F2pr==0):
        F2pr = temp_func
    if(F3pr==0):
        F3pr = temp_func
    if(F4pr==0):
        F4pr = temp_func
    return F(t, y)+(h/2)*(Fpr(t,y))+(h**2/6)*(F2pr(t,y))+(h**3/24)*(F3pr(t,y)) + (h**4/120)*(F4pr(t, y))


def Taylor(y0, h, a, b, plot = 0, F=0, Fpr=0, F2pr=0, F3pr=0, F4pr=0):
    t = [a]
    y = [y0]
    N = int((b-a)/h)
    for i in range(N):
        y.append(y[-1] + h*(T_n(t[-1], y[-1], h, F, Fpr, F2pr, F3pr, F4pr)))
        t.append(t[-1] + h)
    if(plot == 1):
        plt.plot(t, y, label = 'solution')
        plt.legend()
        plt.show()
    return [t, y]



dict1= {
"Euler_key" : ["euler", "eulermethod", "euler'smethod", "forwardeuler"],
"RK_key" : ["rungekutta", "rungekutta4thorder", "rk2", "rk3", "rk4", "rk5",
"rk6", "rungekutta4", "rungekutta2", "rungekutta3", "rungekutta5"],
"AB_key" : ["adambashforth", "ab", "adambashforthexplicit", "adamexplicit"],
"pc_key" : ["predictorcorrector", "pc", "adamspc", "adamsvariablesteppc"
            "adambashforth adammoulton pc, abampc", "variablestepsizepredictorcorrector",
            "pcvariablestepsize", "variablestepsizepc"],
"rkf_key" : ["rungekuttafehlberg", "rkf", "rkfehlberg"],
"taylor_key" : ["taylor", "2ndordertaylor", "secondordertaylor", "3rdordertaylor",
                "thirdordertaylor", "4thordertaylor", "fourthordertaylor",
                "nthordertaylor", "taylorsmethod"],
"meuler_key" : ["modifiedeuler", "meuler", "modifiedforwardeuler", "eulermodified",
                "eulerm", "modifiedeulermethod"]
}






def help(kw = " "):
    if((kw.lower()).replace(" ", "") in dict1["Euler_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes_euler = '''
        Euler's method:
            Function Name: Euler

            Inputs: f: defining function ( from y' = f(t,y) )
                    a: starting Time
                    b: ending Time
                    alpha: Initial Condition
                    h: stepsize
                    plot: Binary input
                        Default plot = 0: No plotting
                                plot = 1: plots the solution

            Outputs: [t, y] where
                y: solution array
                t: time axis
        '''
        print(Notes_euler)

    if((kw.lower()).replace(" ", "") in dict1["meuler_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes_m_euler = '''
        Modified Euler's method:
            Function Name: modified_euler

            Inputs: f: defining function ( from y' = f(t,y) )
                    a: starting Time
                    b: ending Time
                    alpha: Initial Condition
                    h: stepsize
                    plot: Binary input
                        Default plot = 0: No plotting
                                plot = 1: plots the solution

            Outputs: [t, y] where
                y: solution array
                t: time axis
        '''
        print(Notes_m_euler)


    if((kw.lower()).replace(" ", "") in dict1["AB_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes = '''
        Adam Bashforth m step method:
            Function Name: adam_bashforth

            Inputs: f: defining function ( from y' = f(t,y) )
                    a: starting Time
                    b: ending Time
                    alpha: Initial Condition
                    h: stepsize
                    type: # of steps (the 'm' in m-step scheme)
                                        m = 2,3,4 or 5
                    N = (optional) to override the number of iterations
                    default: N = 0 for normal Adam Bashforth explicit
                            reset N to any finite number i to get i iterations.
                                (i >= type-1, because there will be type-1
                                            RK4 runs needed for any AB method)
                    plot: Binary input
                        Default plot = 0: No plotting
                                plot = 1: plots the solution

            Outputs: [t, y]: where
                        t: time axis
                        y: solution array
        '''
        print(Notes)

    if((kw.lower()).replace(" ", "") in dict1["pc_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes = '''
        Does not work on Atom (involves user input - run from console).

         Predictor Corrector scheme (Predictor: Adam Bashforth; Corrector: Adam-Moulton)
             Function Name: predict_correct

             Inputs: f: defining function ( from y' = f(t,y) )
                     alpha: Initial value of the solution
                     a: Initial time
                     b: Final time
                     pred: # steps for predictor (the 'm' in Adam Bashforth m-step)
                     corr: # steps for corrector (the 'm' in Adam Moulton m-step)
                     var_step: Binary input
                        Default var_step = 0: Runs regular predictor corrector
                                var_step = 1: Runs Adams variable step size predictor corrector
                                                (assumes pred = 4, corr = 3)
                     plot: Binary input
                        Default plot = 0: No plotting
                                plot = 1: Plots the solution

                    !! -- Note: depending on user input for var_step, user inputs are
                            requested for h, max h, min h and tol. -- !!

             Outputs:
                    var_step = 0:
                    [t, y]: where
                                t: time axis
                                y: solution array
                    var_step = 1:
                    [t, y, step]: where
                                    t = time axis
                                    y = solution array
                                    step = step-size array

         '''
        print(Notes)

    if((kw.lower()).replace(" ", "") in dict1["RK_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes = '''
        RK for systems (of m equations):
            Note: This function was heavily influenced by the style of Dr. Andasari (BU)
            website: http://people.bu.edu/andasari/courses/numericalpython/python.html

            Function Name: RK4_sys

            Inputs: func: The defining functions written in the following style
                        func(t, y):
                            dy = [0, 0, ..., 0]
                            dy[0] = u1' = f1(t, y[j]); j in 1,2,..., m
                            dy[1] = u2' = f2(t, y[j]); j in 1,2,..., m
                            .
                            .
                            .
                            dy[m] = um' = fm(t, y[j]); j in 1,2,..., m
                            return dy
                    alpha: Array of initial conditions [u1(a), u2(a), ..., um(a)]
                    a: starting time
                    b: ending time
                    h: stepsize
                    plot: Binary input
                       Default plot = 0: No plotting
                               plot = 1: Plots the solution


            Outputs:
                    [tsol, ysol]:
                        ysol: An array of arrays of solutions [u1, u2, ..., um]
                        tsol: Time axis

        =========================== Examples =============================
        Example 1: System of 1 ODE
            Problem:
                y' = (y/t) - (y/t)^2
                t in [1, 4]
                y(1) = 1
                h = 0.1

            #example function to pass into the method:
            def func(t, y):
                dy = np.zeros(len(y))         #len(y) = 1
                dy[0] = (y[0]/t) - (y[0]/t)**2
                return dy

            #declaring variables
            a = 1
            b = 4
            alpha = [1]
            h = 0.1

            #function call
            ODE_Solvers.RK4(func, alpha, a, b, h, plot = 1)
        _______________________________________________________________

        Example 2: higher order ODE (order = 2)
            Problem:
                y'' = t*e**t + 2y' - y
                t in [1, 4]
                y(1) = 1
                y'(1) = 2
                h = 0.1

            {
            #Note: this ode has the following system
            Let u1 = y, u2 = y'
            u1' = u2
            u2' = t*e**t + 2u2 - u1
            with IC
            u1(1) = 1,
            u2(1) = 2
            }

            #example function to pass into the method
            def func(t,y):
                e = np.e
                dy = np.zeros(len(y))           #len(y) = 2, with y[0] = y, and y[1] = y'
                dy[0] = y[1]                    #dy[0] = (u_1)' = y' = u2
                dy[1] = t*e**t + 2*y[1] - y[0]  #dy[1] = (u_2)' = y'' = t*e**t + 2y' - y
                return dy

            #declaring variables
            a = 1
            b = 4
            alpha = [1, 2]
            h = 0.1

            #function call
            ODE_Solvers.RK4(func, alpha, a, b, h, plot = 1)
        ____________________________________________________________

        Example 3: System of first order ODE (Predator-Prey, Volterra Model)
            Problem:
                y1' = k1*y1 - k2*y1*y2
                y2' = k3*y1*y2 - k4*y2
                t in [0, 4]
                y1(0) = 1000
                y2(0) = 500
                h = 0.1

            #example function to pass into the method
            def func(t,y):
                k1 = 3 #change values as needed
                k2 = 0.02 #change values as needed
                k3 = 0.006 #change values as needed
                k4 = 0.5 #change values as needed
                dy = np.zeros(len(y))               #len(y) = 2, with y[0] = u1, and y[1] = u2
                dy[0] = k1*y[0] - k2*y[0]*y[1]      #dy[0] = (u_1)' = y1' = k1*u1 - k2*u1*u2
                dy[1] = k3*y[0]*y[1] - k4*y[1]      #dy[1] = (u_2)' = y2' = k3*u1*u2 - k4*u2
                return dy

            #declaring variables
            a = 0
            b = 4
            alpha = [1000, 500]
            h = 0.1

            #function call
            ODE_Solvers.RK4(func, alpha, a, b, h, plot = 1)

        ___________________________________________________________

        '''
        print(Notes)

    if((kw.lower()).replace(" ", "") in dict1["rkf_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes = '''
        Runge-Kutta Fehlberg:
            Function Name: RKF

            Inputs: f: defining function ( from y' = f(t,y) )
                    a: starting Time
                    b: ending Time
                    alpha: Initial Condition
                    tol: Tolerance
                    hmax: maximum step size
                    hmin: minimum step size
                    plot: Binary input
                        Default plot = 0: No plotting
                                plot = 1: plots the solution

            Outputs: y: solution array
                     t: time array
                     step: stepsize array
        '''
        print(Notes)

    if((kw.lower()).replace(" ", "") in dict1["taylor_key"]):
        print("Method located in numeripy.ODE_solvers \\n")
        Notes = '''
        nth order Taylor method (n = 2, 3, 4, 5):
            Function Name: Taylor

            Inputs: y0: Initial condition
                    h = step size
                    a: starting Time
                    b: ending Time
                    plot: Binary input
                        Default plot = 0: No plotting
                                plot = 1: plots the solution
                    F = f (the defining function y' = f(t, y))
                    Fpr = derivative of f
                    F2pr = second derivative of f
                    F3pr = third derivative of f
                    F4pr = fourth derivative of f

                    Specify only to requirement.
                        -------Example 1:

                        Suppose we want to use 2nd order Taylor. Then define
                        functions f and f' then call function as follows

                        def Taylor(y0, h, a, b, plot = 0, F=f, Fpr=f'):

                        -------Example 2:

                        Suppose we want to use 3rd order Taylor. Then define
                        functions f, f' and f'' then call function as follows

                        def Taylor(y0, h, a, b, plot = 0, F=f, Fpr=f', F2pr = f''):


            Output:  [t, y] where
                        y: solution array
                        t: time array

        '''
        print(Notes)

    return 0



import numpy as np
import matplotlib.pyplot as plt

#---------------------- Runge-Kutta-Fehlberg method ------------------
Notes = '''
Runge-Kutta Fehlberg:
    Function Name: RKF_method

    Inputs: f: defining function ( from y' = f(t,y) )
            a: starting Time
            b: ending Time
            alpha: Initial Condition
            tol: Tolerance
            hmax: maximum step size
            hmin: minimum step size
            plot: Binary input
                Default plot = 0: No plotting
                        plot = 1: plots the solution

    Outputs: y: solution array
             t: time array
             step: stepsize array
'''

def RKF(f, a, b, alpha, tol, hmax, hmin, plot = 0):
    t = a
    w = alpha
    h = hmax
    flag = 1
    time = [a] #time array
    y = [alpha] #solution array
    step = [hmax] #step size array
    while(flag == 1):
        k1 = h*f(t, w)
        k2 = h*f(t + (1/4)*h, w + (1/4)*k1)
        k3 = h*f(t + (3/8)*h, w + (3/32)*k1 + (9/32)*k2)
        k4 = h*f(t + (12/13)*h, w + (1932/2197)*k1 - (7200/2197)*k2 + (7296/2197)*k3)
        k5 = h*f(t + h, w + (439/216)*k1 - (8)*k2 + (3680/513)*k3 - (845/4104)*k4)
        k6 = h*f(t + (1/2)*h, w - (8/27)*k1 + (2)*k2 - (3544/2565)*k3 + (1859/4104)*k4 - (11/40)*k5)
        R = (1/h)*np.abs((1/360)*k1 - (128/4275)*k3 - (2197/75240)*k4 + (1/50)*k5 + (2/55)*k6)
        if(R <= tol):
            t += h
            w += (25/216)*k1 + (1408/2565)*k3 + (2197/4104)*k4 - (1/5)*k5
            time.append(t)
            step.append(h)
            y.append(w)
        delta = 0.84*(tol/R)**(1/4)
        if(delta <= 0.1):
            h = 0.1*h
            step.append(h)
        elif(delta >= 4):
            h = 4*h
            step.append(h)
        else:
            h = delta*h
            step.append(h)
        if(h > hmax):
            h = hmax
            step[-1] = h
        if(t >= b):
            flag = 0
        elif(t + h > b):
            h = b - t
            step[-1] = h
        elif(h<hmin):
            flag = 0
            print("min h exceeded")
            print("completed unsuccessfully")

    if(plot == 1):
        plt.plot(y, time, label = "RKF Solution")
        plt.legend()
        plt.show()

    return y, time, step



import numpy as np
import tabulate
import matplotlib.pyplot as plt

dict = {
"latex_key" : ["latextable", "latex", "table", "tabulate", "latextablecerator",
                "latextableformat", "tabular"],
"plot_key" : ["plot", "graphs", "solutionplot", "odesolution", "plotting"]
}

def help(kw = ""):
    if((kw.lower()).replace(" ", "") in dict["latex_key"]):
        Notes = '''
        Function Name: latexit

        Inputs: cols: # of columns of the table
                headers: list of header names
                *args : all columns as lists (in the same order as headers)

                (optional)
                style:
                  (Default) style = 1: prints latex table with column lines
                            style = 0: prints latex table without column lines


        Outputs: Tabulates given values and prints it in latex table format.
        '''
        print(Notes)

    if((kw.lower()).replace(" ", "") in dict["plot_key"]):
        Notes = '''
        Function Name: plotit

        Inputs: t: the time axis (time array, or x-values array)
                labels: Array of labels of the plots
                *args: all the arrays (to be plotted) in the same order as headers

        Outputs: Plots all the given arrays against the x-values array
                 Prints latex code for including the generated plot
        '''
        print(Notes)

    return 0






#------------------------------------------------------------
#creating latex tables in python!


def latexit(cols, headers, *args, style = 1):
    if(style == 0):
        latexit1(cols, headers, *args)
    if(style == 1):
        latexit2(cols, headers, *args)



def latexit1(cols, headers, *args):
    N = len(args[0])-1
    table = np.zeros((N+1, cols))
    for i in range(N+1):
        temp = []
        for j in range(len(args)):
            temp.append(args[j][i])
        table[i] = temp
    print(tabulate.tabulate(table, \
        headers = headers,  \
        floatfmt=".8f", tablefmt="latex"))

#--------------------------------------------------------------

def latexit2(cols, headers, *args):
    col_str = "c|"
    print("\\begin{center}")
    print("\\begin{tabular}{|%s|%s}"%(col_str, col_str*(cols-1))    )
    print("\\hline")
    string = ""
    for i in range(len(headers)):
        if(i != len(headers) - 1):
            string += "%s   &  "%headers[i]
        else:
            string += "%s \\\ "%headers[i]
    print(string)
    print("\\hline")
    print("\\hline")
    N = len(args[0])-1
    table = np.zeros((N+1, cols))
    for i in range(N+1):
        temp = []
        for j in range(len(args)):
            temp.append(args[j][i])
        table[i] = temp
    for m in range(len(table)):
        string = ""
        for n in range(len(table[m])):
            if(n != len(table[m]) - 1):
                string += "%s   &  "%table[m][n]
            else:
                string += "%s \\\ \\hline"%table[m][n]
        print(string)
    print("\\hline")
    print("\\end{tabular}")
    print("\\end{center}")


def plotit(name, t, label, *args):
    for i in range(len(args)):
        plt.plot(t, args[i], label = label[i])
    plt.legend()
    plt.show()
    print('''
    \\begin{figure}[h]
	\\centering
	\\includegraphics[scale = 0.75]{%s.png}
	\\caption{Plot of the solution}
    \\end{figure}'''%name)



dict2 = {
"gebs_key" : ["gaussianelimination", "gaussianbackwardsubstitution", "gebs",
              "gausianpartialpivoting", "partialpivoting", "gaussianscaledpivoting",
              "scaledpivoting", "gepartialpivoting", "gescaledpivoting"],
"fact_key" : ["lufactorization", "lu", "cholesky", "choleskyfactorization",
              "ldlt", "ldltfactorization"],
"crout_key" : ["crout", "croutfactorization", "croutsolution"],
"iter_key" : ["iterativemethods", "gaussseidel", "jacobi", "sor", "gauss_seidel",
              "gauss-seidel", "jacobiiterativemethod"],
"mult_key" : ["matrixmultiplication", "matmul", "matrixproduct"],
"det_key" : ["det", "determinant"]

}

def help(kw = " "):

  if((kw.lower()).replace(" ", "") in dict2["gebs_key"]):
    print("Method located in numeripy.matrix_methods \\n")
    Notes = '''
    Gaussian Elimination - Backward Substitution
      function name: gaussian_elim

      Inputs: n: Dimension of matrix A
              A: Matrix A (preferably written as a numpy array)
              b: vector b (preferably written as a numpy array)
              ver: Ternary input
                  Default ver = 0: Ordinary Gaussian Elimination
                          ver = 1: Gaussian Elimination with partial pivoting
                          ver = 2: Gaussian Elimination with scaled pivoting

      Outputs: x: solution vector (returned as a numpy array)
    '''
    print(Notes)


  if((kw.lower()).replace(" ", "") in dict2["fact_key"]):
    print("Method located in numeripy.matrix_methods \\n")
    Notes = '''
    Factorization methods:
      function name: matrix_fact

      Inputs: n: Dimension of matrix A
              A: Matrix A (preferably written as a numpy array)
              type:
                Default type = "lu": LU factorization
                        type = "ldlt": LDLt factorization
                        type = "cholesky": Cholesky factorization
      Outputs:
          type = "lu": [L, U]
          type = "cholesky": L
          type = "ldlt": [L, D]
    '''
    print(Notes)

  if((kw.lower()).replace(" ", "") in dict2["crout_key"]):
    print("Method located in numeripy.matrix_methods \\n")
    Notes = '''
    Crout mathod
      function name: crout

      Inputs: n: Dimension of matrix A
              A: Matrix A (preferably written as a numpy array)
              b: vector b (preferably written as a numpy array)

      Outputs: x (as a numpy array. Solution to Ax = B)
    '''
    print(Notes)

  if((kw.lower()).replace(" ", "") in dict2["iter_key"]):
    print("Method located in numeripy.matrix_methods \\n")
    Notes =  '''
    Iterative matrix methods
      function name: mat_iter

      Inputs: A: Matrix A (preferably written as a numpy array)
              b: vector b (preferably written as a numpy array)
              xo: initial x value (preferably written as a numpy array)
              N: maximum number of iteration
              w: default w = 0: For any method
                         w = any number: for SOR method
              type:
                type = "jacobi": Jacobi iterative method
                type = "gauss-seidel": Gauss Seidel iterative method
                type = "sor": SOR iterative method
    '''
    print(Notes)

  if((kw.lower()).replace(" ", "") in dict2["mult_key"]):
    print("Method located in numeripy.matrix_methods \\n")
    Notes = '''
    Matrix multiplication
      function name: Matmul

      Inputs: A: Matrix A (preferably written as a numpy array)
              B: Matrix B (preferably written as a numpy array)

      Outputs: C = AB the product matrix (as a numpy array)
    '''
    print(Notes)

  if((kw.lower()).replace(" ", "") in dict2["det_key"]):
    print("Method located in numeripy.matrix_methods \\n")
    Notes = '''
    Determinant
      function name: determinant

      Inputs: A: Matrix A (preferably written as a numpy array)

      Outputs: det(A) the determinant of A
    '''
    print(Notes)
  return 0



import numpy as np
import matplotlib.pyplot as plt

Notes_m_euler = '''
Modified Euler's method:
    Function Name: modified_euler

    Inputs: f: defining function ( from y' = f(t,y) )
            a: starting Time
            b: ending Time
            alpha: Initial Condition
            h: stepsize
            plot: Binary input
                Default plot = 0: No plotting
                        plot = 1: plots the solution

    Outputs: [t, y] where
        y: solution array
        t: time axis
'''
def modified_euler(f, a, b, alpha, h, plot = 0):
    t = [a]
    y = [alpha]
    N = int((b-a)/h)
    for i in range(N):
        val = f(t[-1],y[-1])
        y.append(y[-1]  + (h/2)*(val + f(t[-1]+h, y[-1]+h*val)))
        t.append(t[-1] + h)

    if(plot == 1):
        plt.plot(t, y, label = "Modified Euler Solution")
        plt.legend()
        plt.show()

    return [t, y]



import numpy as np

def mat_iter(A, b, xo, N, tol, w = 0, type = " "):
    if((type.lower()).replace(" ", "") == "jacobi"):
        return Jacobi(A, b, xo, N, tol)
    if((type.lower()).replace(" ", "") == "gaussseidel"):
        return Gauss_Seidel(A, b, xo, N, tol)
    if((type.lower()).replace(" ", "") == "sor"):
        return SOR(A, b, xo, w, tol, N)

def Jacobi(A, b, xo, N, tol):
    n = len(b)
    k = 1
    x = np.zeros(n)
    while(k<=N):
        for i in range(0,n):
            sum1 = 0
            for j in range(0,n):
                if(i != j):
                    sum1+=A[i][j]*xo[j]
            x[i] = (1/A[i][i])*(-sum1 + b[i])
        if(np.linalg.norm(x-xo, np.inf) <= tol):
            return x
        k+=1
        for i in range(0, n):
            xo[i] = x[i]
    print("Max iterations exceeded")
    return x

def Gauss_Seidel(A, b, xo, N, tol):
    n = len(b)
    k = 1
    x = np.zeros(n)

    while(k<=N):
        for i in range(n):
            sum1 = 0
            sum2 = 0
            for j in range(i):
                sum1 += A[i][j]*x[j]
            for m in range(i+1, n):
                sum2 +=A[i][m]*xo[m]
            x[i] = (1/A[i][i])*(-1*sum1-sum2+b[i])
        if(np.linalg.norm(x-xo)<=tol):
            return x
        k+=1
        for p in range(n):
            xo[p] = x[p]
    print("Max iterations exceeded")
    return x

def SOR(A, b, xo, w, tol, N):
    n = len(b)
    x = np.zeros(n)
    k = 1
    while(k<=N):
        for i in range(0,n):
            sum1 = 0
            sum2 = 0
            for j in range(0,i):
                sum1 +=A[i][j]*x[j]
            for j in range(i+1,n):
                sum2+=A[i][j]*xo[j]
            x[i] = (1-w)*xo[i] + (w/A[i][i])*(-sum1-sum2+b[i])
        if(np.linalg.norm(x-xo, np.inf) < tol):
            return x
        k+=1
        for i in range(0,n):
            xo[i] = x[i]
        print(x)
    print("Maximum iteration exceeded")
    return x



from distutils.core import setup
setup(
  name = 'numeripy',
  packages = ['numeripy', 'numeripy.ODE_solvers', 'numeripy.matrix_methods'],
  version = '0.1.1',
  license='MIT',
  description = 'Python package for numerical analysis. Includes a variety of niche ode solvers and iterative matrix methods (typically encountered in a standard senior year course in numerical analysis)',   # Give a short description about your library
  author = 'Arun Suresh',
  author_email = 'ab251098@gmail.com',
  url = 'https://github.com/asuresh213/numeripy',
  download_url = 'https://github.com/asuresh213/numeripy/archive/v_01.tar.gz',
  keywords = ['ODE Solvers', 'Numerical', 'Numerical Analysis', 'Optimization',
              'Matrix Methods', 'ODE', 'Differential Equations', 'Iterative matrix methods'],
  install_requires=[
          'numpy',
          'tabulate',
          'matplotlib',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
